"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import approaching_number_limit


class LimitNumber(CloudFormationLintRule):
    """Check maximum Mapping limit"""

    id = 'I7010'
    shortdesc = 'Mapping limit'
    description = (
        'Check the number of Mappings in the template is approaching the upper limit'
    )
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html'
    tags = ['mappings', 'limits']

    def match(self, cfn):
        return approaching_number_limit(cfn, 'Mappings')



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import number_limit


class LimitNumber(CloudFormationLintRule):
    """Check if maximum Resource limit is exceeded"""

    id = 'E3010'
    shortdesc = 'Resource limit not exceeded'
    description = (
        'Check the number of Resources in the template is less than the upper limit'
    )
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html'
    tags = ['resources', 'limits']

    def match(self, cfn):
        return number_limit(cfn, 'Resources')



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class Config(CloudFormationLintRule):
    """Check if Metadata configuration is properly configured"""

    id = 'E4002'
    shortdesc = 'Validate the configuration of the Metadata section'
    description = 'Validates that Metadata section is an object and has no null values'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/metadata-section-structure.html'
    tags = ['metadata']

    def _check_object(self, obj, path):
        results = []
        if isinstance(obj, (dict)):
            for k, v in obj.items():
                results.extend(self._check_object(v, path + [k]))
        if isinstance(obj, (list)):
            for i, v in enumerate(obj):
                results.extend(self._check_object(v, path + [i]))
        if obj is None:
            message = 'Metadata value cannot be null'
            results.append(RuleMatch(path, message.format(message)))

        return results

    def match(self, cfn):
        """Check CloudFormation Metadata Interface Configuration"""

        matches = []

        metadata_obj = cfn.template.get('Metadata', {})
        if metadata_obj is None:
            message = 'Metadata value has to be an object'
            matches.append(RuleMatch(['Metadata'], message.format(message)))

        if metadata_obj:
            if isinstance(metadata_obj, dict):
                matches.extend(self._check_object(metadata_obj, ['Metadata']))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch
from cfnlint.helpers import LIMITS


class LimitDescription(CloudFormationLintRule):
    """Check Template Description Size"""

    id = 'E1003'
    shortdesc = 'Template description limit'
    description = (
        'Check if the size of the template description is less than the upper limit'
    )
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html'
    tags = ['description', 'limits']

    def match(self, cfn):
        matches = []
        description = cfn.template.get('Description', '')
        if isinstance(description, str):
            if len(description) > LIMITS['template']['description']:
                message = (
                    'The template description ({0} bytes) exceeds the limit ({1} bytes)'
                )
                matches.append(
                    RuleMatch(
                        ['Description'],
                        message.format(
                            len(description), LIMITS['template']['description']
                        ),
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import REGISTRY_SCHEMAS
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch
import cfnlint.helpers


class Configuration(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = 'E3001'
    shortdesc = 'Basic CloudFormation Resource Check'
    description = (
        'Making sure the basic CloudFormation resources are properly configured'
    )
    source_url = 'https://github.com/aws-cloudformation/cfn-python-lint'
    tags = ['resources']

    def _check_resource(self, cfn, resource_name, resource_values):
        """Check Resource"""

        valid_attributes = [
            'Condition',
            'CreationPolicy',
            'DeletionPolicy',
            'DependsOn',
            'Metadata',
            'Properties',
            'Type',
            'UpdatePolicy',
            'UpdateReplacePolicy',
        ]

        valid_custom_attributes = [
            'Condition',
            'DeletionPolicy',
            'DependsOn',
            'Metadata',
            'Properties',
            'Type',
            'UpdateReplacePolicy',
            'Version',
        ]

        matches = []
        if not isinstance(resource_values, dict):
            message = 'Resource not properly configured at {0}'
            matches.append(
                RuleMatch(['Resources', resource_name], message.format(resource_name))
            )
            return matches

        # validate condition is a string
        condition = resource_values.get('Condition', '')
        if not isinstance(condition, str):
            message = 'Condition for resource {0} should be a string'
            matches.append(
                RuleMatch(
                    ['Resources', resource_name, 'Condition'],
                    message.format(resource_name),
                )
            )

        resource_type = resource_values.get('Type', '')
        if not isinstance(resource_type, str):
            message = 'Type has to be a string at {0}'
            matches.append(
                RuleMatch(
                    ['Resources', resource_name],
                    message.format('/'.join(['Resources', resource_name])),
                )
            )
            return matches

        # Type is valid continue analysis
        if (
            resource_type.startswith('Custom::')
            or resource_type == 'AWS::CloudFormation::CustomResource'
        ):
            check_attributes = valid_custom_attributes
        else:
            check_attributes = valid_attributes

        for property_key, _ in resource_values.items():
            if property_key not in check_attributes:
                message = 'Invalid resource attribute {0} for resource {1}'
                matches.append(
                    RuleMatch(
                        ['Resources', resource_name, property_key],
                        message.format(property_key, resource_name),
                    )
                )

        if not resource_type:
            message = 'Type not defined for resource {0}'
            matches.append(
                RuleMatch(['Resources', resource_name], message.format(resource_name))
            )
        elif not isinstance(resource_type, str):
            message = 'Type has to be a string at {0}'
            matches.append(
                RuleMatch(
                    ['Resources', resource_name],
                    message.format('/'.join(['Resources', resource_name])),
                )
            )
        else:
            self.logger.debug('Check resource types by region...')
            for region, specs in cfnlint.helpers.RESOURCE_SPECS.items():
                if region in cfn.regions:
                    if resource_type not in specs[
                        'ResourceTypes'
                    ] and resource_type not in [
                        s['typeName'] for s in REGISTRY_SCHEMAS
                    ]:
                        if not resource_type.startswith(
                            ('Custom::', 'AWS::Serverless::')
                        ) and not resource_type.endswith('::MODULE'):
                            message = 'Invalid or unsupported Type {0} for resource {1} in {2}'
                            matches.append(
                                RuleMatch(
                                    ['Resources', resource_name, 'Type'],
                                    message.format(
                                        resource_type, resource_name, region
                                    ),
                                )
                            )

        if 'Properties' not in resource_values:
            resource_spec = cfnlint.helpers.RESOURCE_SPECS[cfn.regions[0]]
            if resource_type in resource_spec['ResourceTypes']:
                properties_spec = resource_spec['ResourceTypes'][resource_type][
                    'Properties'
                ]
                # pylint: disable=len-as-condition
                if len(properties_spec) > 0:
                    required = 0
                    for _, property_spec in properties_spec.items():
                        if property_spec.get('Required', False):
                            required += 1
                    if required > 0:
                        if (
                            resource_type == 'AWS::CloudFormation::WaitCondition'
                            and 'CreationPolicy' in resource_values.keys()
                        ):
                            self.logger.debug(
                                'Exception to required properties section as CreationPolicy is defined.'
                            )
                        else:
                            message = 'Properties not defined for resource {0}'
                            matches.append(
                                RuleMatch(
                                    ['Resources', resource_name],
                                    message.format(resource_name),
                                )
                            )

        return matches

    def match(self, cfn):
        matches = []

        resources = cfn.template.get('Resources', {})
        if not isinstance(resources, dict):
            message = 'Resource not properly configured'
            matches.append(RuleMatch(['Resources'], message))
        else:
            for resource_name, resource_values in cfn.template.get(
                'Resources', {}
            ).items():
                self.logger.debug(
                    'Validating resource %s base configuration', resource_name
                )
                matches.extend(
                    self._check_resource(cfn, resource_name, resource_values)
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class PreviousGenerationInstanceType(CloudFormationLintRule):
    id = 'I3100'
    shortdesc = 'Checks for legacy instance type generations'
    description = 'New instance type generations increase performance and decrease cost'
    source_url = 'https://aws.amazon.com/ec2/previous-generation/'
    tags = ['resources', 'ec2', 'rds', 'elasticcache', 'elasticsearch']

    def match(self, cfn):
        matches = []
        for resource_type, property_type in [
            ('AWS::AutoScaling::LaunchConfiguration', 'InstanceType'),
            ('AWS::EC2::CapacityReservation', 'InstanceType'),
            ('AWS::EC2::Host', 'InstanceType'),
            ('AWS::EC2::Instance', 'InstanceType'),
            ('AWS::RDS::DBInstance', 'DBInstanceClass'),
            ('AWS::ElastiCache::CacheCluster', 'CacheNodeType'),
            ('AWS::ElastiCache::GlobalReplicationGroup', 'CacheNodeType'),
            ('AWS::ElastiCache::ReplicationGroup', 'CacheNodeType'),
        ]:
            for resource_name, resource in cfn.get_resources([resource_type]).items():
                if isinstance(
                    resource.get('Properties', {}).get(property_type, ''), str
                ):
                    if re.search(
                        r'([cmr][1-3]|cc2|cg1|cr1|g2|hi1|hs1|i2|t1)\.',
                        resource.get('Properties', {}).get(property_type, ''),
                    ):
                        matches.append(
                            RuleMatch(
                                [
                                    'Resources',
                                    resource_name,
                                    'Properties',
                                    property_type,
                                ],
                                'Upgrade previous generation instance type: '
                                + resource.get('Properties').get(property_type),
                            )
                        )

        for resource_type, top_level_property_type, property_type in [
            (
                'AWS::EC2::EC2Fleet',
                'FleetLaunchTemplateOverridesRequest',
                'InstanceType',
            ),
            ('AWS::EC2::LaunchTemplate', 'LaunchTemplateData', 'InstanceType'),
            ('AWS::EC2::SpotFleet', 'SpotFleetLaunchSpecification', 'InstanceType'),
            ('AWS::OpenSearchService::Domain', 'ClusterConfig', 'InstanceType'),
            (
                'AWS::Elasticsearch::Domain',
                'ElasticsearchClusterConfig',
                'InstanceType',
            ),
        ]:
            for resource_name, resource in cfn.get_resources([resource_type]).items():
                if isinstance(
                    resource.get('Properties', {})
                    .get(top_level_property_type, {})
                    .get(property_type, ''),
                    str,
                ):
                    if re.search(
                        r'([cmr][1-3]|cc2|cg1|cr1|g2|hi1|hs1|i2|t1)\.',
                        resource.get('Properties', {})
                        .get(top_level_property_type, {})
                        .get(property_type, ''),
                    ):
                        matches.append(
                            RuleMatch(
                                [
                                    'Resources',
                                    resource_name,
                                    'Properties',
                                    top_level_property_type,
                                    property_type,
                                ],
                                'Upgrade previous generation instance type: '
                                + resource.get('Properties')
                                .get(top_level_property_type)
                                .get(property_type),
                            )
                        )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class Configuration(CloudFormationLintRule):
    """Check if Mappings are configured correctly"""

    id = 'E7001'
    shortdesc = 'Mappings are appropriately configured'
    description = 'Check if Mappings are properly configured'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/mappings-section-structure.html'
    tags = ['mappings']

    def match(self, cfn):
        matches = []

        valid_map_types = (str, list, int, float)

        mappings = cfn.template.get('Mappings', {})
        if mappings:
            for mapname, mapobj in mappings.items():
                if not isinstance(mapobj, dict):
                    message = 'Mapping {0} has invalid property'
                    matches.append(
                        RuleMatch(['Mappings', mapname], message.format(mapname))
                    )
                else:
                    for firstkey in mapobj:
                        firstkeyobj = mapobj[firstkey]
                        if not isinstance(firstkeyobj, dict):
                            message = 'Mapping {0} has invalid property at {1}'
                            matches.append(
                                RuleMatch(
                                    ['Mappings', mapname, firstkey],
                                    message.format(mapname, firstkeyobj),
                                )
                            )
                        else:
                            for secondkey in firstkeyobj:
                                if not isinstance(
                                    firstkeyobj[secondkey], valid_map_types
                                ):
                                    message = 'Mapping {0} has invalid property at {1}'
                                    matches.append(
                                        RuleMatch(
                                            ['Mappings', mapname, firstkey, secondkey],
                                            message.format(mapname, secondkey),
                                        )
                                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class InterfaceParameterExists(CloudFormationLintRule):
    """Check if Metadata Interface parameters exist"""

    id = 'W4001'
    shortdesc = 'Metadata Interface parameters exist'
    description = 'Metadata Interface parameters actually exist'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-cloudformation-interface.html'
    tags = ['metadata']

    valid_keys = ['ParameterGroups', 'ParameterLabels']

    def match(self, cfn):
        """Check CloudFormation Metadata Parameters Exist"""

        matches = []

        strinterface = 'AWS::CloudFormation::Interface'
        parameters = cfn.get_parameter_names()
        metadata_obj = cfn.template.get('Metadata', {})
        if metadata_obj:
            interfaces = metadata_obj.get(strinterface, {})
            if isinstance(interfaces, dict):
                # Check Parameter Group Parameters
                paramgroups = interfaces.get('ParameterGroups', [])
                if isinstance(paramgroups, list):
                    for index, value in enumerate(paramgroups):
                        if 'Parameters' in value:
                            for paramindex, paramvalue in enumerate(
                                value['Parameters']
                            ):
                                if paramvalue not in parameters:
                                    message = 'Metadata Interface parameter doesn\'t exist {0}'
                                    matches.append(
                                        RuleMatch(
                                            [
                                                'Metadata',
                                                strinterface,
                                                'ParameterGroups',
                                                index,
                                                'Parameters',
                                                paramindex,
                                            ],
                                            message.format(paramvalue),
                                        )
                                    )
                paramlabels = interfaces.get('ParameterLabels', {})
                if isinstance(paramlabels, dict):
                    for param in paramlabels:
                        if param not in parameters:
                            message = 'Metadata Interface parameter doesn\'t exist {0}'
                            matches.append(
                                RuleMatch(
                                    [
                                        'Metadata',
                                        strinterface,
                                        'ParameterLabels',
                                        param,
                                    ],
                                    message.format(param),
                                )
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import os
from pathlib import Path
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch
from cfnlint.helpers import LIMITS


class LimitSize(CloudFormationLintRule):
    """Check Template Size"""

    id = 'I1002'
    shortdesc = 'Template size limit'
    description = 'Check the size of the template is approaching the upper limit'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html'
    tags = ['limits']

    def match(self, cfn):
        matches = []
        # Only check if the file exists. The template could be passed in using stdIn
        if cfn.filename:
            if Path(cfn.filename).is_file():
                statinfo = os.stat(cfn.filename)
                if (
                    LIMITS['threshold'] * LIMITS['template']['body']
                    < statinfo.st_size
                    <= LIMITS['template']['body']
                ):
                    message = 'The template file size ({0} bytes) is approaching the limit ({1} bytes)'
                    matches.append(
                        RuleMatch(
                            ['Template'],
                            message.format(
                                statinfo.st_size, LIMITS['template']['body']
                            ),
                        )
                    )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import valid_snapshot_types
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class DeletionPolicy(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = 'E3035'
    shortdesc = 'Check DeletionPolicy values for Resources'
    description = 'Check that the DeletionPolicy values are valid'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html'
    tags = ['resources', 'deletionpolicy']

    def check_value(self, key, path, res_type, has_lang_exten_transform):
        """Check resource names for DeletionPolicy"""
        matches = []

        valid_values = ['Delete', 'Retain', 'Snapshot']

        supported_functions = ['Fn::FindInMap', 'Fn::If', 'Ref']

        supported_functions_joined = ', '.join(supported_functions)

        if has_lang_exten_transform and isinstance(key, dict):
            if len(key) == 1:
                for index_key, _ in key.items():
                    if index_key not in supported_functions:
                        message = (
                            'DeletionPolicy only supports one of the '
                            + supported_functions_joined
                            + ' intrinsic functions for {0}'
                        )
                        matches.append(
                            RuleMatch(path, message.format('/'.join(map(str, path))))
                        )
            else:
                message = 'DeletionPolicy should have one mapping for {0}'
                matches.append(
                    RuleMatch(path, message.format('/'.join(map(str, path))))
                )
        else:
            if not isinstance(key, (str)):
                message = 'DeletionPolicy values should be of string at {0}'
                matches.append(
                    RuleMatch(path, message.format('/'.join(map(str, path))))
                )
                return matches
            if key not in valid_values:
                message = 'DeletionPolicy should be only one of {0} at {1}'
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            ', '.join(map(str, valid_values)), '/'.join(map(str, path))
                        ),
                    )
                )
            if key == 'Snapshot' and res_type not in valid_snapshot_types:
                message = (
                    'DeletionPolicy cannot be Snapshot for resources of type {0} at {1}'
                )
                matches.append(
                    RuleMatch(path, message.format(res_type, '/'.join(map(str, path))))
                )

        return matches

    def match(self, cfn):
        matches = []

        resources = cfn.get_resources()

        for resource_name, resource_values in resources.items():
            deletion_policies = resource_values.get('DeletionPolicy')
            if deletion_policies:
                path = ['Resources', resource_name, 'DeletionPolicy']
                res_type = resource_values.get('Type')
                self.logger.debug(
                    'Validating DeletionPolicy for %s base configuration', resource_name
                )
                if isinstance(deletion_policies, list):
                    message = 'Only one DeletionPolicy allowed per resource at {0}'
                    matches.append(
                        RuleMatch(path, message.format('/'.join(map(str, path))))
                    )
                else:
                    has_lang_exten_transform = cfn.has_language_extensions_transform()
                    matches.extend(
                        self.check_value(
                            deletion_policies, path, res_type, has_lang_exten_transform
                        )
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules.resources.properties.BasedOnValue import BasedOnValue


class UnwantedBasedOnValue(BasedOnValue):
    """Check Unwanted Properties are supplied when another property has a certain value"""

    id = 'E3018'
    shortdesc = 'Property is unwanted based on another properties value'
    description = (
        'When certain properties have a certain value it results in other properties not being needed. '
        'This rule will validate those unwanted properties are not specified when those values are supplied'
    )
    tags = ['resources']
    spec_type = 'UnwantedProperties'
    message = 'should not be specified'

    def _check_prop(self, prop, scenario):
        return prop in scenario.get('Object')



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class Elb(CloudFormationLintRule):
    """Check if Elb Resource Properties"""

    id = 'E2503'
    shortdesc = 'Resource ELB Properties'
    description = 'See if Elb Resource Properties are set correctly \
HTTPS has certificate HTTP has no certificate'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-elb-listener.html'
    tags = ['properties', 'elb']

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types = ['AWS::ElasticLoadBalancingV2::LoadBalancer']

    def check_protocol_value(self, value, path, **kwargs):
        """
        Check Protocol Value
        """
        matches = []
        if isinstance(value, str):
            if value.upper() not in kwargs['accepted_protocols']:
                message = 'Protocol must be {0} is invalid at {1}'
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            (', '.join(kwargs['accepted_protocols'])),
                            ('/'.join(map(str, path))),
                        ),
                    )
                )
            elif value.upper() in kwargs['certificate_protocols']:
                if not kwargs['certificates']:
                    message = (
                        'Certificates should be specified when using HTTPS for {0}'
                    )
                    matches.append(
                        RuleMatch(path, message.format(('/'.join(map(str, path)))))
                    )

        return matches

    def get_loadbalancer_type(self, properties):
        """Check if type is application"""
        elb_type = properties.get('Type', 'application')
        if isinstance(elb_type, str):
            if elb_type == 'application':
                return 'application'
            return 'network'
        return None

    def check_alb_subnets(self, properties, path, scenario):
        """Validate at least two subnets with ALBs"""
        matches = []
        if self.get_loadbalancer_type(properties) == 'application':
            subnets = properties.get('Subnets')
            if isinstance(subnets, list):
                if len(subnets) < 2:
                    if scenario:
                        message = 'You must specify at least two Subnets for load balancers with type "application" {0}'
                        scenario_text = ' and '.join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        matches.append(RuleMatch(path, message.format(scenario_text)))
                    else:
                        matches.append(
                            RuleMatch(
                                path[:] + ['Subnets'],
                                'You must specify at least two Subnets for load balancers with type "application"',
                            )
                        )
            subnet_mappings = properties.get('SubnetMappings')
            if isinstance(subnet_mappings, list):
                if len(subnet_mappings) < 2:
                    if scenario:
                        message = 'You must specify at least two SubnetMappings for load balancers with type "application" {0}'
                        scenario_text = ' and '.join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        matches.append(RuleMatch(path, message.format(scenario_text)))
                    else:
                        matches.append(
                            RuleMatch(
                                path[:] + ['SubnetMappings'],
                                'You must specify at least two SubnetMappings for load balancers with type "application"',
                            )
                        )

        return matches

    def match(self, cfn):
        """Check ELB Resource Parameters"""

        matches = []

        results = cfn.get_resource_properties(['AWS::ElasticLoadBalancingV2::Listener'])
        for result in results:
            matches.extend(
                cfn.check_value(
                    result['Value'],
                    'Protocol',
                    result['Path'],
                    check_value=self.check_protocol_value,
                    accepted_protocols=[
                        'GENEVE',
                        'HTTP',
                        'HTTPS',
                        'TCP',
                        'TCP_UDP',
                        'TLS',
                        'UDP',
                    ],
                    certificate_protocols=['HTTPS', 'TLS'],
                    certificates=result['Value'].get('Certificates'),
                )
            )

        results = cfn.get_resource_properties(
            ['AWS::ElasticLoadBalancing::LoadBalancer', 'Listeners']
        )
        for result in results:
            if isinstance(result['Value'], list):
                for index, listener in enumerate(result['Value']):
                    matches.extend(
                        cfn.check_value(
                            listener,
                            'Protocol',
                            result['Path'] + [index],
                            check_value=self.check_protocol_value,
                            accepted_protocols=['HTTP', 'HTTPS', 'TCP', 'SSL'],
                            certificate_protocols=['HTTPS', 'SSL'],
                            certificates=listener.get('SSLCertificateId'),
                        )
                    )

        return matches

    def match_resource_properties(self, resource_properties, _, path, cfn):
        """Check Load Balancers"""
        matches = []

        # Play out conditions to determine the relationship between property values
        scenarios = cfn.get_object_without_nested_conditions(resource_properties, path)
        for scenario in scenarios:
            properties = scenario.get('Object')
            if self.get_loadbalancer_type(properties) == 'network':
                if properties.get('SecurityGroups'):
                    if scenario.get('Scenario'):
                        scenario_text = ' and '.join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.get('Scenario').items()
                            ]
                        )
                        message = f'Security groups are not supported for load balancers with type "network" {scenario_text}'
                        matches.append(RuleMatch(path, message))
                    else:
                        path = path + ['SecurityGroups']
                        matches.append(
                            RuleMatch(
                                path,
                                'Security groups are not supported for load balancers with type "network"',
                            )
                        )

            matches.extend(
                self.check_alb_subnets(properties, path, scenario.get('Scenario'))
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.languageExtensions import LanguageExtensions
from cfnlint.rules import RuleMatch


class Length(CloudFormationLintRule):
    """Check if Length values are correct"""

    id = 'E1030'
    shortdesc = 'Length validation of parameters'
    description = 'Making sure Fn::Length is configured correctly'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-length.html'
    tags = ['functions', 'length']

    def match(self, cfn):
        has_language_extensions_transform = cfn.has_language_extensions_transform()
        intrinsic_function = 'Fn::Length'
        matches = []

        fn_length_objects = cfn.search_deep_keys(intrinsic_function)
        for fn_length_object in fn_length_objects:
            tree = fn_length_object[:-1]
            LanguageExtensions.validate_transform_is_declared(
                self,
                has_language_extensions_transform,
                matches,
                tree,
                intrinsic_function,
            )
            self.validate_type(fn_length_object, matches, tree)
            self.validate_ref(fn_length_object, matches, tree, cfn)
        return matches

    def validate_ref(self, fn_length_object, matches, tree, cfn):
        fn_length_value = fn_length_object[-1]
        if isinstance(fn_length_value, dict):
            if len(fn_length_value.keys()) != 1 or (
                list(fn_length_value.keys())[0] not in ['Ref', 'Fn::Split']
            ):
                self.addMatch(
                    matches,
                    tree,
                    'Fn::Length expects either an array, a Ref to an array or Fn::Split, '
                    'but found unexpected object under Fn::Length at {0}',
                )
                return

            if 'Ref' in fn_length_value:
                if fn_length_value['Ref'] not in cfn.get_parameter_names():
                    self.addMatch(
                        matches,
                        tree,
                        'Fn::Length can only reference list parameters at {0}',
                    )
                else:
                    referenced_parameter = cfn.get_parameters().get(
                        fn_length_value['Ref']
                    )
                    parameter_type = referenced_parameter.get('Type')
                    if 'List' not in parameter_type:
                        self.addMatch(
                            matches,
                            tree,
                            'Fn::Length can only reference list parameters at {0}',
                        )

    def addMatch(self, matches, tree, message):
        matches.append(RuleMatch(tree[:], message.format('/'.join(map(str, tree)))))

    def validate_type(self, fn_length_object, matches, tree):
        fn_length_value = fn_length_object[-1]
        if not isinstance(fn_length_value, dict) and not isinstance(
            fn_length_value, list
        ):
            message = 'Fn::Length needs a list or a reference to a list at {0}'
            matches.append(RuleMatch(tree[:], message.format('/'.join(map(str, tree)))))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from datetime import datetime
from cfnlint.rules.resources.lmbd.DeprecatedRuntime import DeprecatedRuntime
from cfnlint.rules import RuleMatch


class DeprecatedRuntimeEol(DeprecatedRuntime):
    """Check if EOL Lambda Function Runtimes are used"""

    id = 'W2531'
    shortdesc = 'Check if EOL Lambda Function Runtimes are used'
    description = (
        'Check if an EOL Lambda Runtime is specified and give a warning if used. '
    )
    source_url = (
        'https://docs.aws.amazon.com/lambda/latest/dg/runtime-support-policy.html'
    )
    tags = ['resources', 'lambda', 'runtime']

    def check_runtime(self, runtime_value, path):
        """Check if the given runtime is valid"""
        matches = []

        runtime = self.deprecated_runtimes.get(runtime_value)
        if runtime:
            if (
                datetime.strptime(runtime['eol'], '%Y-%m-%d') < self.current_date
                and datetime.strptime(runtime['deprecated'], '%Y-%m-%d')
                > self.current_date
            ):
                message = 'EOL runtime ({0}) specified. Runtime is EOL since {1} and updating will be disabled at {2}. Please consider updating to {3}'
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            runtime_value,
                            runtime['eol'],
                            runtime['deprecated'],
                            runtime['successor'],
                        ),
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class Ref(CloudFormationLintRule):
    """Check if Ref value is a string"""

    id = 'E1020'
    shortdesc = 'Ref validation of value'
    description = (
        'Making sure the Ref has a String value (no other functions are supported)'
    )
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html'
    tags = ['functions', 'ref']

    def match(self, cfn):
        matches = []

        ref_objs = cfn.search_deep_keys('Ref')
        for ref_obj in ref_objs:
            value = ref_obj[-1]
            if not isinstance(value, (str)):
                message = 'Ref can only be a string for {0}'
                matches.append(
                    RuleMatch(
                        ref_obj[:-1], message.format('/'.join(map(str, ref_obj[:-1])))
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules import RuleMatch


class Configuration(CloudFormationLintRule):
    """Check if Parameters are configured correctly"""

    id = 'E2001'
    shortdesc = 'Parameters have appropriate properties'
    description = 'Making sure the parameters are properly configured'
    source_url = 'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html'
    tags = ['parameters']

    valid_keys = {
        'AllowedPattern': {'Type': 'String'},
        'AllowedValues': {
            'Type': 'List',
            'ItemType': 'String',
        },
        'ConstraintDescription': {'Type': 'String'},
        'Default': {'Type': 'String'},
        'Description': {'Type': 'String'},
        'MaxLength': {
            'Type': 'Integer',
            'ValidForTypes': [
                'String',
                'AWS::EC2::AvailabilityZone::Name',
                'AWS::EC2::Image::Id',
                'AWS::EC2::Instance::Id',
                'AWS::EC2::KeyPair::KeyName',
                'AWS::EC2::SecurityGroup::GroupName',
                'AWS::EC2::SecurityGroup::Id',
                'AWS::EC2::Subnet::Id',
                'AWS::EC2::Volume::Id',
                'AWS::EC2::VPC::Id',
                'AWS::Route53::HostedZone::Id',
            ],
        },
        'MaxValue': {'Type': 'Integer', 'ValidForTypes': ['Number']},
        'MinLength': {
            'Type': 'Integer',
            'ValidForTypes': [
                'String',
                'AWS::EC2::AvailabilityZone::Name',
                'AWS::EC2::Image::Id',
                'AWS::EC2::Instance::Id',
                'AWS::EC2::KeyPair::KeyName',
                'AWS::EC2::SecurityGroup::GroupName',
                'AWS::EC2::SecurityGroup::Id',
                'AWS::EC2::Subnet::Id',
                'AWS::EC2::Volume::Id',
                'AWS::EC2::VPC::Id',
                'AWS::Route53::HostedZone::Id',
            ],
        },
        'MinValue': {'Type': 'Integer', 'ValidForTypes': ['Number']},
        'NoEcho': {'Type': 'Boolean'},
        'Type': {'Type': 'String'},
    }

    required_keys = ['Type']

    def check_type(self, value, path, props):
        """Check the type and handle recursion with lists"""
        results = []
        prop_type = props.get('Type')
        if value is None:
            message = (
                f'Property {"/".join(map(str, path))} should be of type {prop_type}'
            )
            results.append(RuleMatch(path, message))
            return results
        try:
            if prop_type in ['List']:
                if isinstance(value, list):
                    for i, item in enumerate(value):
                        results.extend(
                            self.check_type(
                                item, path[:] + [i], {'Type': props.get('ItemType')}
                            )
                        )
                else:
                    message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                    results.append(RuleMatch(path, message))
            if prop_type in ['String']:
                if isinstance(value, (dict, list)):
                    message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                    results.append(RuleMatch(path, message))
                str(value)
            elif prop_type in ['Boolean']:
                if not isinstance(value, bool):
                    if value not in ['True', 'true', 'False', 'false']:
                        message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                        results.append(RuleMatch(path, message))
            elif prop_type in ['Integer']:
                if isinstance(value, bool):
                    message = f'Property {"/".join(map(str, path))} should be of type {prop_type}'
                    results.append(RuleMatch(path, message))
                else:  # has to be a Double
                    int(value)
        except Exception:  # pylint: disable=W0703
            message = (
                f'Property {"/".join(map(str, path))} should be of type {prop_type}'
            )
            results.append(
                RuleMatch(
                    path,
                    message,
                )
            )

        return results

    def match(self, cfn):
        matches = []

        for paramname, paramvalue in cfn.get_parameters().items():
            if isinstance(paramvalue, dict):
                for propname, propvalue in paramvalue.items():
                    if propname not in self.valid_keys:
                        message = 'Parameter {0} has invalid property {1}'
                        matches.append(
                            RuleMatch(
                                ['Parameters', paramname, propname],
                                message.format(paramname, propname),
                            )
                        )
                    else:
                        props = self.valid_keys.get(propname)
                        prop_path = ['Parameters', paramname, propname]
                        matches.extend(self.check_type(propvalue, prop_path, props))
                        # Check that the property is needed for the current type
                        valid_for = props.get('ValidForTypes')
                        if valid_for is not None:
                            if paramvalue.get('Type') not in valid_for:
                                message = 'Parameter {0} has property {1} which is only valid for {2}'
                                matches.append(
                                    RuleMatch(
                                        ['Parameters', paramname, propname],
                                        message.format(paramname, propname, valid_for),
                                    )
                                )

                for reqname in self.required_keys:
                    if reqname not in paramvalue.keys():
                        message = 'Parameter {0} is missing required property {1}'
                        matches.append(
                            RuleMatch(
                                ['Parameters', paramname],
                                message.format(paramname, reqname),
                            )
                        )
            else:
                message = 'Parameter {0} is not an object'
                matches.append(
                    RuleMatch(
                        ['Parameters', paramname], message.format(paramname, reqname)
                    )
                )

        return matches
